@@ -1,5 +1,5 @@
 //
-// $Id: dnstracer.c,v 1.21 2002/01/22 04:17:49 mavetju Exp $
+// $Id: dnstracer.c,v 1.25 2002/01/25 12:05:46 mavetju Exp $
 //
 
 //
@@ -29,23 +29,34 @@
 // OF THE POSSIBILITY OF SUCH DAMAGE.
 //
 
+#ifdef WIN32
+    #include <winsock.h>
+    #include <io.h>
+    #include "getopt.h"
+    // To reduce the amount of #ifdef's, these two are defined.
+    // rand()/srand() is ISO C89 naming but it's obsoleted according
+    // to the man-page.
+    #define random	rand
+    #define srandom	srand
+#else
+    #include <unistd.h>
+    #include <sys/time.h>
+    #include <sys/param.h>
+    #include <sys/socket.h>
+    #include <netinet/in.h>
+    #include <arpa/inet.h>
+    #include <arpa/nameser.h>
+    #include <netdb.h>
+    #include <resolv.h>
+#endif
+
 #include <sys/types.h>
-#include <sys/param.h>
-#include <sys/socket.h>
-#include <netinet/in.h>
-#include <arpa/inet.h>
-#include <arpa/nameser.h>
-#include <netdb.h>
-
-#include "dnstracer_broken.h"
-
-#include <sys/time.h>
 #include <stdio.h>
 #include <string.h>
 #include <stdlib.h>
 #include <time.h>
-#include <unistd.h>
-#include <resolv.h>
+
+#include "dnstracer_broken.h"
 
 #define DEFAULT_RETRIES	3
 #define DEFAULT_TIMEOUT	15
@@ -913,26 +924,6 @@ int is_busy(char *server) {
 // The core of this program
 //
 
-char *mygethostbyname(char *name) {
-    struct hostent *h;
-    static char s[NS_MAXDNAME];
-
-    //
-    // Convert a hostname into an dotted IP address
-    //
-
-    if ((h=gethostbyname(name))==NULL)
-	return NULL;
-
-    sprintf(s,"%hu.%hu.%hu.%hu",
-	(unsigned char)h->h_addr_list[0][0],
-	(unsigned char)h->h_addr_list[0][1],
-	(unsigned char)h->h_addr_list[0][2],
-	(unsigned char)h->h_addr_list[0][3]);
-
-    return s;
-}
-
 int create_session(int dnssocket,char *host,
 		    char *server_ip,char *server_name,
 		    int depth,char *prefix,int last) {
@@ -966,7 +957,7 @@ int create_session(int dnssocket,char *host,
     // records in it might be unreachable.
     //
     if (is_busy(server_ip)==1) {
-	printf("RECURSION DETECTED ");
+	printf("Lame server ");
 	fflush(stdout);
 	return 0;
     }
@@ -1062,7 +1053,7 @@ int create_session(int dnssocket,char *host,
 	//
 	rradd=session->additional;
 	while (rradd!=NULL) {
-	   if (strcmp(rradd->domainname,rrauth->data_name)==0)
+	   if (strcmp(printablename(rradd->domainname,1),rrauth->data_name)==0)
 		found++;
 	    rradd=rradd->next;
 	}
@@ -1073,43 +1064,87 @@ int create_session(int dnssocket,char *host,
 	    // Find the first IP address
 	    //
 	    while (rradd!=NULL) {
-		if (strcmp(rradd->domainname,rrauth->data_name)==0)
+		if (strcmp(printablename(rradd->domainname,1),
+			   rrauth->data_name)==0)
 		    break;
 		rradd=rradd->next;
 	    }
 
-	    printf("\n");
-
 	    //
 	    // Prepare the graphs in front of the servers name
 	    //
 	    sprintf(s,"%s%c%s",
 		prefix,last==1?' ':'|',depth==0?"":"     ");
 
-	    if (rradd!=NULL) {
-		strcpy(nextserver_name,printablename(rradd->domainname,1));
-		strcpy(nextserver_ip,rradd->data_name);
-	    } else {
-		char *s;
-
-		strcpy(nextserver_name,rrauth->data_name);
-		if ((s=mygethostbyname(nextserver_name))==NULL)
-		    nextserver_ip[0]=0;
-		else
-		    strcpy(nextserver_ip,s);
-	    }
-
 	    //
 	    // Recurse into this server...
 	    //
-	    retval+=create_session(dnssocket,host,nextserver_ip,nextserver_name,
+	    if (rradd!=NULL) {
+		// This is easy, we got the IP address in the additional
+		// section. Don't worry about additional records with the
+		// same name, they're done automaticly after this one.
+
+		printf("\n");
+
+		strcpy(nextserver_name,printablename(rradd->domainname,1));
+		strcpy(nextserver_ip,rradd->data_name);
+		retval+=create_session(dnssocket,host,
+			    nextserver_ip,nextserver_name,
+			    depth+1,s,(rrauth->next==NULL && found<=1)?1:0);
+	    } else {
+		struct hostent *h;
+
+		strcpy(nextserver_name,rrauth->data_name);
+
+		if ((h=gethostbyname(nextserver_name))==NULL) {
+		    //
+		    // No IP address was found for this hostname.
+		    // Just call the function and let them print
+		    // an error.
+		    //
+		    printf("\n");
+		    nextserver_ip[0]=0;
+		    retval+=create_session(dnssocket,host,
+			nextserver_ip,nextserver_name,
 			depth+1,s,(rrauth->next==NULL && found<=1)?1:0);
+		} else {
+		    int count,i;
+		    char **addr_list;
+
+		    //
+		    // One or more IP address were found. Go through all
+		    // of them (make sure they're saved before calling
+		    // gethostbyname() again).
+		    //
+		    count=0;
+		    while (h->h_addr_list[count]!=NULL) count++;
+
+		    addr_list=(char **)calloc(count,sizeof(char *));
+		    for (i=0;i<count;i++) {
+			addr_list[i]=(char *)calloc(1,h->h_length);
+			memcpy(addr_list[i],h->h_addr_list[i],h->h_length);
+		    }
+
+		    for (i=0;i<count;i++) {
+			printf("\n");
+			sprintf(nextserver_ip,"%hu.%hu.%hu.%hu",
+			    (unsigned char)addr_list[i][0],
+			    (unsigned char)addr_list[i][1],
+			    (unsigned char)addr_list[i][2],
+			    (unsigned char)addr_list[i][3]);
+
+			retval+=create_session(dnssocket,host,
+			    nextserver_ip,nextserver_name,
+			    depth+1,s,(rrauth->next==NULL && found<=1)?1:0);
+		    }
+		}
+	    }
 
 	    //
 	    // If there are no more IP addresses, then do the next
 	    // authorityRR.
 	    //
-	    if (--found==0)
+	    if (--found<=0)
 		break;
 	    if (rradd!=NULL)
 		rradd=rradd->next;
@@ -1127,6 +1162,13 @@ int create_session(int dnssocket,char *host,
     return retval;
 }
 
+/*****************************************************************************/
+
+//
+// Startup, usage and win32-initialization
+// Win32-initialization by Mike Black <mblack@csihq.com>
+//
+
 void usage(void) {
     fprintf(stderr,
 	"Usage: dnstracer [options] [host]\n"
@@ -1142,14 +1184,27 @@ void usage(void) {
     exit(1);
 }
 
+#ifdef WIN32
+int wsockinit(void) {
+    WSADATA wsaData;		/* Structure for WinSock setup communication */
+
+    if (WSAStartup(MAKEWORD(2, 0), &wsaData) != 0) { /* Load Winsock 2.0 DLL */
+	fprintf(stderr, "WSAStartup() failed");
+	exit(1);
+    }
+    return 1;
+}
+#endif
+
+
 int main(int argc,char **argv) {
     int		dnssocket;
     char	ch;
-    char *	server_name;
-    char *	server_ip;
+    char *	server_name="127.0.0.1";
+    char *	server_ip="127.0.0.1";
     char	ipaddress[NS_MAXDNAME];
 
-
+#ifndef WIN32
     //
     // Get the first nameserver from /etc/resolv.conf
     //
@@ -1158,13 +1213,15 @@ int main(int argc,char **argv) {
     if (!(_res.options & RES_INIT))
 	res_init();
 
-    if (_res.nscount<1) {
-	server_name="127.0.0.1";
-	server_ip="127.0.0.1";
-    } else {
+    if (_res.nscount>0) {
 	server_ip=strdup(inet_ntoa(_res.nsaddr_list[0].sin_addr));
 	server_name=strdup(server_ip);
     }
+#endif
+
+#ifdef WIN32
+    wsockinit();
+#endif
 
     while ((ch=getopt(argc,argv,"coq:r:s:t:v"))!=-1) {
 	switch (ch) {
@@ -1238,11 +1295,18 @@ int main(int argc,char **argv) {
 
     srandom(time(NULL));
 
-    if ((server_ip=mygethostbyname(server_name))==NULL) {
-	fprintf(stderr,"Cannot find IP address for %s\n",server_name);
-	return 1;
+    {
+	struct hostent *h;
+	if ((h=gethostbyname(server_name))==NULL) {
+	    fprintf(stderr,"Cannot find IP address for %s\n",server_name);
+	    return 1;
+	}
+	sprintf(ipaddress,"%hu.%hu.%hu.%hu",
+	    (unsigned char)h->h_addr_list[0][0],
+	    (unsigned char)h->h_addr_list[0][1],
+	    (unsigned char)h->h_addr_list[0][2],
+	    (unsigned char)h->h_addr_list[0][3]);
     }
-    strcpy(ipaddress,server_ip);
 
     create_session(dnssocket,argv[0],ipaddress,server_name,0,"",1);
 
